using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;

using TAPI;
using Terraria;

using FearTheReaper;
using FearTheReaper.Items;

namespace FearTheReaper.NPCs {
	public class ReaperMobNPC : ModNPC {
		/** Aura 'very far' radius (in blocks). */
		public readonly static int AuraVeryFar = 52;
		/** Aura 'far' radius (in blocks). */
		public readonly static int AuraFar = 36;
		/** Aura 'near' radius (in blocks). */
		public readonly static int AuraNear = 20;
		/** Aura 'very near' radius (in blocks). */
		public readonly static int AuraVeryNear = 10;


		/**	Time (in milliseconds) between each reaper rage reduction (-1). */
		private long ReaperRageCooldownTime = 16000L;	// 16 seconds

		/**	Timer used for counting off whole seconds. */
		private Stopwatch WholeSecondClock = new Stopwatch();


		/**	Timer for velocity clamping (speed constraining) skip (ie rage bursts, knockback, etc). */
		private Stopwatch VelocityClampSkipClock = new Stopwatch();

		/**	Ambient sound loop instance. */
		private SoundEffectInstance AmbienceInstance = null;
		/**	Ambient sound loop timer. TODO: Find a better way to loop audio! */
		private Stopwatch AmbienceClock = new Stopwatch();
		/**	Timer used for calming rage. */
		private Stopwatch PatienceClock = new Stopwatch();
		

		/**	Your angel of death awaits! */
		private static Dictionary<ReaperMobNPC, FTR_Player> PlayerTargets = new Dictionary<ReaperMobNPC, FTR_Player>();


		/**	Initial knockback resist number of this reaper. */
		private float BaseKnockbackResist = 1f;


		/**	Measures total time spent haunting. */
		private Stopwatch HauntClock = new Stopwatch();


		private bool hasStealth = true;
		/**	Indicates if our reaper has its stealth on. */
		public bool HasStealth { get { return hasStealth; } }


		private bool isFleeing = false;
		/** Indicates our reaper has decided to flee. */
		public bool IsFleeing { get { return isFleeing; } }

		private int DefaultAlpha;
		private float CurrentAlpha;

		private bool _IsInited = false;


		private static float CurrentMusicVolumeScale = 1f;



		/**	Gets a player's reaper. */
		public static ReaperMobNPC GetMyReaper( FTR_Player ftr_player ) {
			ReaperMobNPC ftr_reaper = null;

			foreach( KeyValuePair<ReaperMobNPC, FTR_Player> entry in PlayerTargets ) {
				ftr_reaper = entry.Key;
				FTR_Player map_ftr_player = entry.Value;
				if( map_ftr_player == ftr_player ) { break; }
			}

			if( ftr_reaper == null ) { return null; }

			int reaper_type = NPCDef.byName["FearTheReaper:ReaperMobNPC"].type;
			if( !ftr_reaper.npc.active || ftr_reaper.npc.type != reaper_type ) {
				PlayerTargets.Remove( ftr_reaper );
				ftr_reaper = null;
			}

			return ftr_reaper;
		}


		/**	Removes a player's reaper. */
		public static bool ReapMyReaper( FTR_Player ftr_player, bool kill_only ) {
			ReaperMobNPC ftr_reaper = GetMyReaper( ftr_player );
			if( ftr_reaper == null ) { return false; }

			if( kill_only ) {
				ftr_reaper.npc.life = 0;
			} else {
				RoboUtils.DestroyNPC( ftr_reaper.npc );
			}

			PlayerTargets.Remove( ftr_reaper );

			return true;
		}


		
		public static bool CanSpawnIntoPool( int x, int y, System.Collections.Generic.IList<int> pool, Player player ) {
		//public override bool CanSpawn( int x, int y, int type, Player player ) {
			FTR_Player ftr_player = player.GetSubClass<FTR_Player>();
			if( ftr_player == null ) {
RoboUtils.MsgAll( "Can't spawn: No player.", RoboUtils.MsgType.Debug );
				return false; }

			if( !ftr_player.AllowsReaper() ) {
RoboUtils.MsgAll( "Can't spawn: Reaper not allowed. Player: \""+player.name+"\"", RoboUtils.MsgType.Debug );return false; }

			if( !RoboUtils.IsRNGDisabled ) {
				int base_spawn_odds = 42;
				int rage_odds_drop = ftr_player.ReaperRage;	// (int)((float)ftr_player.ReaperRage * 1.5f);
				int linger_odds_drop = Math.Min( (int)((float)ftr_player.LingerMeasure / 3.3f), (base_spawn_odds - 9) );
				int spawn_odds = Math.Max( base_spawn_odds - (rage_odds_drop + linger_odds_drop), 1 );
				if( Main.rand.Next( spawn_odds ) != 0 ) {
RoboUtils.MsgAll( "Can't spawn: RNG. "+base_spawn_odds+" - (R="+rage_odds_drop+" + L="+linger_odds_drop+") = "+spawn_odds, RoboUtils.MsgType.Debug );
					return false; }
			}
			
			// Be sure our player isn't already being chased
			ReaperMobNPC ftr_reaper = GetMyReaper(ftr_player);
			if( ftr_reaper != null ) {
RoboUtils.MsgAll( "Can't spawn: Already being chased. (who:"+ftr_reaper.npc.whoAmI+", where:"+ftr_reaper.npc.position.X+","+ftr_reaper.npc.position.Y+")", RoboUtils.MsgType.Debug );
				return false; }

			bool is_on_surface = (double)player.position.Y < (Main.worldSurface * 16.0);
			bool is_in_hell = (double)player.position.Y > (Main.hellLayer * 16.0);

			// No boss?
			if( RoboUtils.IsCurrentPlayerNearBoss() ) {
RoboUtils.MsgAll( "Can't spawn: Boss near.", RoboUtils.MsgType.Debug );return false; }

			// Player isn't in underworld?
			if( is_in_hell ) {
RoboUtils.MsgAll( "Can't spawn: In hell.", RoboUtils.MsgType.Debug );return false; }

			// No eclipse/pumpkin/frost moon (while on surface)?
			if( is_on_surface && ( Main.eclipse || Main.snowMoon || Main.pumpkinMoon ) ) {
RoboUtils.MsgAll( "Can't spawn: Moon event.", RoboUtils.MsgType.Debug );return false; }

			// No (non-blood moon) npc town?
			if( is_on_surface && player.townNPCs >= 1f && !Main.bloodMoon ) {
RoboUtils.MsgAll( "Can't spawn: In town.", RoboUtils.MsgType.Debug );
				return false; }

			return true;
		}



		////////////////////////////////////////////////////////////////
		////////////////////////////////////////////////////////////////

		public override void OnSpawn() {
			DefaultAlpha = npc.alpha;
			CurrentAlpha = (float)DefaultAlpha;

			//CurrentMusicVolumeScale = WavebankDef.fade[WavebankDef.current];
			CurrentMusicVolumeScale = 1f;

			// Is this needed?!
			//NetMessage.SendData( 23, -1, -1, "", npc.whoAmI, 0f, 0f, 0f, 0f );

			HauntClock.Start();
			BaseKnockbackResist = Math.Max( npc.knockBackResist, 0.1f );
		}



		//public override bool? CanHitNPC( Player player ) {	// Doesn't work?
//            can_dmg = CanDamageMe( player );
//RoboUtils.MsgAll( "Can damage? " + can_dmg, RoboUtils.MsgType.Debug );
		//    return can_dmg;
		//}

		public override void DamageNPC( Projectile projectile, int hitDir, ref int damage, ref float knockback, ref bool crit, ref float critMult ) {
			if( projectile.player != null ) {
				DamageMe( projectile.player, ref damage, ref crit );
			}
		}

		public override void DamageNPC( Player player, int hitDir, ref int damage, ref float knockback, ref bool crit, ref float critMult ) {
			DamageMe( player, ref damage, ref crit );
		}


		public override void HitEffect( int hit_direction, double damage, bool is_dead ) {
			if( RoboUtils.IsClient() ) {
				bool dramatic = damage > 20 || is_dead;
				int particles = dramatic ? 25 : 5;
				int particle_range_width = dramatic ? npc.width * 2 : npc.width;
				int particle_range_height = dramatic ? npc.height * 2 : npc.height;

				for( int m = 0; m < particles; m++ ) {
					float veloc_x = npc.velocity.X * 0.1f;
					float veloc_y = npc.velocity.Y * 0.1f;
					float scale = dramatic ? 2f : 1f;
					int dust_id = Dust.NewDust( npc.position, particle_range_width, particle_range_height, 15, veloc_x, veloc_y, 200, Color.White, scale );
					if( dust_id < Main.dust.Length ) { Main.dust[dust_id].noGravity = true; }
				}
			}
		}


		public override void PostNPCLoot() {
			if( (float)Main.rand.NextDouble() < 0.02f ) {	//2% chance
				int token = ItemDef.byName["FearTheReaper:FerryTokenAccessoryItem"].type;
				Item.NewItem( (int)npc.position.X, (int)npc.position.Y, npc.width, npc.height, token, 1, false, 0 );
			}
		}


		public override void PostAI() {
			UpdateMe();
		}

		////////////////////////////////////////////////////////////////
		////////////////////////////////////////////////////////////////



		public bool CanDamageMe( Player player ) {
			FTR_Player ftr_player = player.GetSubClass<FTR_Player>();
			bool can_dmg = ftr_player.ReaperRage < 10;
			return can_dmg;
		}


		public void DamageMe( Player player, ref int damage, ref bool crit ) {
			// Knockback should bypass clamping
			if( !VelocityClampSkipClock.IsRunning ) {
				VelocityClampSkipClock.Reset();
				VelocityClampSkipClock.Start();
			}

			FTR_Player ftr_player = player.GetSubClass<FTR_Player>();

			if( HasStealth ) {
				hasStealth = false;
RoboUtils.MsgAll( "Stealth off (shot) " + npc.alpha, RoboUtils.MsgType.Debug );
			}

			ftr_player.ReaperRage++;
RoboUtils.MsgAll( "Grrrr " + ftr_player.ReaperRage, RoboUtils.MsgType.Debug );

			// TODO: Figure out how to actually stop the damage
			if( !CanDamageMe(player) ) {
				Rectangle npc_rect = new Rectangle( (int)npc.position.X, (int)npc.position.Y, npc.width, npc.height );
				CombatText.NewText( npc_rect, new Color( 100, 255, 100, 255 ), string.Concat(1), false, false );
				npc.life += 1;
			}
		}


		private bool CanFlee() {
			if( IsFleeing ) { return false; }

			// Reaper always ought to get SOME haunting time in...
			if( HauntClock.ElapsedMilliseconds <= 24000L ) { return false; }	// 24 seconds

			// No player target? Give it a chance.
			if( !PlayerTargets.ContainsKey( this ) ) { return false; }
			FTR_Player ftr_player = PlayerTargets[this];

			float distance = Vector2.Distance( ftr_player.player.position, npc.position ) / 16f;
			float progress_scale = (float)ftr_player.ComputeProgressMeasure() / (float)FTR_Player.AdvancedProgressMilestone;
			float min_threshold = (float)AuraFar;	// (AuraNear + AuraFar) / 2f;
			float max_threshold = (float)(AuraVeryFar + (AuraNear / 2));
			float threshold = MathHelper.Clamp( (float)(AuraVeryFar * progress_scale), min_threshold, max_threshold );

			// Far enough to flee?
			if( distance <= threshold ) { return false; }

			// Rage decreases chance of flee
			int odds = 3 + ftr_player.ReaperRage;
			int rand = Main.rand.Next( odds );
			if( rand != 0 ) {
				_DespawnAttempts++;
				RoboUtils.MsgAll( "Could not despawn. (odds: " + odds + ", _DespawnAttempts: " + _DespawnAttempts + ")", RoboUtils.MsgType.Debug );
				return false;
			}
			RoboUtils.MsgAll( "I got a fever. (odds: " + odds + ", _DespawnAttempts: " + _DespawnAttempts + ")", RoboUtils.MsgType.Debug );
			_DespawnAttempts = 0;

			return true;
		}
		private int _DespawnAttempts = 0;

		public void FleeMe() {
			if( IsFleeing ) { return; }
			isFleeing = true;

			if( PlayerTargets.ContainsKey( this ) ) {
				PlayerTargets[this].ReaperRage += 1;
			}
			RoboUtils.MsgAll( " cya " + PlayerTargets[this].ReaperRage, RoboUtils.MsgType.Debug );
		}


		private void CalmMe() {
			if( !PlayerTargets.ContainsKey( this ) ) { return; }
			FTR_Player ftr_player = PlayerTargets[this];

			float progress = (float)ftr_player.ComputeProgressMeasure() / (float)FTR_Player.AdvancedProgressMilestone;
			int gtfo = ftr_player.ReaperRage + 4 + (int)(progress * 4f);

			// Force despawn more if we're early into the game (and at least 24 seconds into haunting)
			if( HauntClock.ElapsedMilliseconds >= 24000L && Main.rand.Next( gtfo ) == 0 ) {
				RoboUtils.MsgAll( "le sigh (" + ftr_player.ReaperRage + ", gtfo:" + gtfo + ")", RoboUtils.MsgType.Debug );
				FleeMe();
			}
				// Otherwise, simply calm the reaper
			else {
				ftr_player.ReaperRage--;
				RoboUtils.MsgAll( "sigh (rage:" + ftr_player.ReaperRage + ", gtfo:" + gtfo + ")", RoboUtils.MsgType.Debug );
			}
		}




		private void UpdateMe() {
			if( !WholeSecondClock.IsRunning ) { WholeSecondClock.Start(); }

			bool whole_second = WholeSecondClock.ElapsedMilliseconds >= 1000L;
			if( whole_second ) {
				WholeSecondClock.Reset();
				WholeSecondClock.Start();
			}

			RegisterPlayerTarget();

			// TODO: Find a better way to complete initialization
            if( !_IsInited ) {
                _IsInited = true;

                // On each new spawn, increase rage
                if( PlayerTargets.ContainsKey( this ) ) {
					//PlayerTargets[this].ReaperRage += 1;
RoboUtils.MsgAll( "*** Needs more cowbell! (rage: " + PlayerTargets[this].ReaperRage + ")", RoboUtils.MsgType.Debug );
                }
            }

			if( whole_second ) {
				if( CanFlee() ) {
					FleeMe();
				}
			}

			if( UpdateExistantState() ) {
				UpdateSpeed( whole_second );
				UpdateRage();
				UpdateAmbience();
				UpdateMusic( whole_second );
				UpdateAura();
				UpdateLightsOut();
				UpdateBuffCleanse();
				UpdateStealth();
				if( whole_second ) { UpdateDeathAura(); }
			}
		}


		/** Registers this reaper's respective player target.	TODO: Address possible redundancy? */
		private void RegisterPlayerTarget() {
			Player player = Main.player[npc.target];
			if( player == null || !player.active ) {
				PlayerTargets.Remove( this );
RoboUtils.MsgAll( "RegisterPlayerTarget failed: Player not in game", RoboUtils.MsgType.Debug );
				return;
			}

			FTR_Player ftr_player = player.GetSubClass<FTR_Player>();
			if( ftr_player == null ) {
RoboUtils.MsgAll( "RegisterPlayerTarget failed: No FTR player", RoboUtils.MsgType.Debug );
				return; }

			// Garbage collect any old reapers
			List<ReaperMobNPC> old_reaps = new List<ReaperMobNPC>();
			foreach( KeyValuePair<ReaperMobNPC, FTR_Player> entry in PlayerTargets ) {
				ReaperMobNPC reaper = entry.Key;
				if( reaper.npc == null || !reaper.npc.active ) {
					old_reaps.Add( reaper );
				}
			}
			foreach( ReaperMobNPC reaper in old_reaps ) {
				PlayerTargets.Remove( reaper );
			}

			PlayerTargets[this] = ftr_player;
		}


		/**	Updates whether reaper exists or not. */
		private bool UpdateExistantState() {
			bool is_existant = npc.active;

			if( is_existant && IsFleeing ) {
				// Nothing to target!
				if( !PlayerTargets.ContainsKey( this ) ) { return true; }
				FTR_Player ftr_player = PlayerTargets[this];
				Player player = ftr_player.player;

				//float progress_scale = (float)ftr_player.ComputeProgressMeasure() / (float)FTR_Player.AdvancedProgressMilestone;

				float distance = Vector2.Distance( player.position, npc.position ) / 16f;
				float threshold = AuraVeryFar;

				if( distance > threshold ) {
					RoboUtils.DestroyNPC( npc );
					is_existant = false;
				}
			}

			return is_existant;
		}



		/**	Applies reaper 'clinginess' upon player (scales with distance and 'ReaperRage'). */
		private static void ApplyPlayerClingVelocity( FTR_Player ftr_player, NPC reaper ) {
			Player player = ftr_player.player;

			float reaper_distance = Vector2.Distance( player.position, reaper.position ) / 16f;
			float min_distance_threshold = AuraFar;	//(AuraFar + AuraNear) / 2;
			float max_distance_threshold = (AuraNear + AuraVeryNear) / 2;

			// Clinginess falls off when near
			float distance_scale = (reaper_distance - max_distance_threshold) / (min_distance_threshold - max_distance_threshold);
			distance_scale = MathHelper.Clamp( distance_scale, 0f, 1f );

			//// Player 'progress' adds to cling
			//float progress_scale = (float)ftr_player.ComputeProgressMeasure() / (float)FTR_Player.AdvancedProgressMilestone;
			//float progress_cling = progress_scale * distance_scale;
			
			// Minimum cling
			float scaled_min_cling = 0.33f * distance_scale;

			// Cling added from rage
			float scaled_rage = (float)ftr_player.ReaperRage * distance_scale;
			float rage_cling = scaled_rage / 33f;

			float cling = Math.Min( scaled_min_cling + rage_cling /*+ progress_cling*/, 1f );
			if( cling == 0 ) { return; }

			if( (reaper.position.X < player.position.X && player.velocity.X > 0) || (reaper.position.X > player.position.X && player.velocity.X < 0) ) {
				reaper.velocity.X += player.velocity.X * cling;
			}
			if( (reaper.position.Y < player.position.Y && player.velocity.Y > 0) || (reaper.position.Y > player.position.Y && player.velocity.Y < 0) ) {
				reaper.velocity.Y += player.velocity.Y * cling;
			}
		}

		/**	Clamps reaper velocity to a maximum (raised by player's 'ReaperRage' and progress). */
		private static void ApplyVelocityClamp( FTR_Player ftr_player, NPC reaper, bool is_whole_second=false ) {
			float veloc_max_base = 1.1f;

			// Widen clamp with rage
			float rage_veloc_max_add = (float)ftr_player.ReaperRage / 5f;	// 6.6f too slow? 3.3f too fast?

			// Widen clamp with progress
			float progress_scale = (float)ftr_player.ComputeProgressMeasure() / (float)FTR_Player.AdvancedProgressMilestone;

			// Tighten clamp if in blocks
			float friction_scale = 1f;
			int x_beg = (int)(reaper.position.X / 16f);
			int x_end = (int)((reaper.position.X + reaper.width) / 16f);
			int y_beg = (int)(reaper.position.Y / 16f);
			int y_end = (int)((reaper.position.Y + reaper.height) / 16f);
			if( Collision.SolidTiles( x_beg, x_end, y_beg, y_end ) ) {
				friction_scale = 0.65f;
if( is_whole_second ) { RoboUtils.MsgAll( "reaper is in matter", RoboUtils.MsgType.Debug ); }
			}

			float max_veloc = (veloc_max_base + rage_veloc_max_add + (progress_scale * 2f)) * friction_scale;

			float reaper_velocity = (float)Math.Sqrt( (reaper.velocity.X * reaper.velocity.X) + (reaper.velocity.Y * reaper.velocity.Y) );
			if( reaper_velocity > max_veloc ) {
				float dampen_scale = max_veloc / reaper_velocity;

				reaper.velocity.X *= dampen_scale;
				reaper.velocity.Y *= dampen_scale;
			}
		}


		private void UpdateSpeed( bool is_whole_second=false ) {
			// Nothing to target; we're free!
			if( !PlayerTargets.ContainsKey(this) ) { return; }
			FTR_Player ftr_player = PlayerTargets[this];
			Player player = ftr_player.player;

			if( IsFleeing ) {
				if( npc.position.Y > player.position.Y ) {
					npc.velocity.Y = Math.Max( npc.velocity.Y, 1.2f ) + 0.1f;
				} else {
					npc.velocity.Y = Math.Min( npc.velocity.Y, -1.2f ) - 0.1f;
				}
			}

			// Skip velocity altering?
			if( VelocityClampSkipClock.IsRunning ) {
				if( VelocityClampSkipClock.ElapsedMilliseconds >= 750L ) {	// 0.75 second
					VelocityClampSkipClock.Reset();
					VelocityClampSkipClock.Stop();
				}
			}
			else if( !IsFleeing ) {
				// Add velocity for target player cling behavior
				ApplyPlayerClingVelocity( ftr_player, npc );

				// Clamp by stat modifiers
				ApplyVelocityClamp( ftr_player, npc, is_whole_second );
			}
		}


		private void UpdateRage() {
			if( !PlayerTargets.ContainsKey( this ) ) { return; }
			FTR_Player ftr_player = PlayerTargets[this];

			// Calm rage with prolonged chase
			if( PatienceClock.IsRunning ) {
				if( PatienceClock.ElapsedMilliseconds >= ReaperRageCooldownTime ) {
					PatienceClock.Reset();
					PatienceClock.Start();

					CalmMe();
				}
			} else {
				PatienceClock.Reset();
				PatienceClock.Start();
			}

			// Scale knockback resist with rage
			float rage_scale = (float)ftr_player.ReaperRage / 10f;
			float resist = BaseKnockbackResist - (rage_scale * BaseKnockbackResist);
			npc.knockBackResist = resist;

			// Scale reaper's, uh, scale with rage
			npc.scale = 1f + (rage_scale * 0.85f);
		}


		private void UpdateAmbience() {
			// Terminate sound if npc inactive
			if( !npc.active ) {
				AmbienceClock.Stop();
				if( AmbienceInstance != null ) { AmbienceInstance.Stop(); }
				AmbienceInstance = null;
				return;
			}

			// Stream ambient sound
			if( AmbienceInstance != null ) {
				if( AmbienceClock.IsRunning && AmbienceClock.ElapsedMilliseconds >= 3000L ) {	// 3 seconds
					if( AmbienceInstance.State != SoundState.Playing ) {
						AmbienceInstance.Play();
						AmbienceClock.Reset();
						AmbienceClock.Start();
					}
				}
			}
			// Begin ambient sound
			else {
				AmbienceInstance = Main.PlaySound( "FearTheReaper:Wind", npc.position.X, npc.position.Y );

				//AmbienceInstance.IsLooped = true;	// <- Doesn't work?
				if( !AmbienceClock.IsRunning ) { AmbienceClock.Start(); }
			}

			RoboUtils.UpdateSoundForCurrentListener( AmbienceInstance, (int)npc.position.X, (int)npc.position.Y, 0.65f );
		}


		private void UpdateMusic( bool is_whole_second ) {
			Player curr_player = Main.localPlayer;
			if( curr_player == null || !curr_player.active ) { return; }
			if( !WavebankDef.fade.ContainsKey( WavebankDef.current ) ) { return; }

			float distance = Vector2.Distance( curr_player.position, npc.position ) / 16f;

			// Fade music based on distance
			float volume_scale = MathHelper.Clamp( ((distance - (float)AuraNear) / (float)AuraVeryFar), 0f, 1f );
			if( CurrentMusicVolumeScale > volume_scale && !IsFleeing && !curr_player.dead ) {
				CurrentMusicVolumeScale = volume_scale;
			} else {
				CurrentMusicVolumeScale = Math.Min( CurrentMusicVolumeScale + 0.002f, 1f );
			}
			WavebankDef.fade[WavebankDef.current] = CurrentMusicVolumeScale;
if( is_whole_second ) { RoboUtils.MsgAll( "fade:" + CurrentMusicVolumeScale, RoboUtils.MsgType.Debug ); }
		}



		private void UpdateAura() {
			Player curr_player = Main.localPlayer;
			if( curr_player == null || !curr_player.active ) { return; }
			
			if( !curr_player.dead ) {
				float distance = Vector2.Distance( curr_player.position, npc.position ) / 16f;

				// No teleporting
				if( distance <= AuraVeryFar ) {
					curr_player.AddBuff( "FearTheReaper:AnchoredBuff", 2 );
					//curr_player.AddBuff( "Chaos State", 2 );	// Redundant with Anchored
				}
				// Weakened attacks, slower movement, no health regen
				if( distance <= AuraFar ) {
					curr_player.AddBuff( "Chilled", 2 );
					curr_player.AddBuff( "Weak", 2 );
					curr_player.AddBuff( "Bleeding", 2 );
				}
				// Darkened vision, no magic, yet slower movement, no potion healing
				if( distance <= AuraNear ) {
					//curr_player.AddBuff( "Slow", 2 );	// Removed as of v1.4
					curr_player.AddBuff( "Silenced", 2 );
					curr_player.AddBuff( "Darkness", 2 );
					curr_player.AddBuff( "Potion Sickness", 4 );
					//curr_player.AddBuff( "Confused", 2 );	// Fun, but pretty over powered
				}
				// Much darkened vision, no item use at all, a small speed boost, damage over time
				if( distance <= AuraVeryNear ) {
					curr_player.AddBuff( "Blackout", 2 );
					curr_player.AddBuff( "Cursed", 2 );
					curr_player.AddBuff( "Panic!", 2 );
					curr_player.AddBuff( "Suffocation", 2 );
				}
			}
		}

		private void UpdateLightsOut() {
			Rectangle near = new Rectangle( (int)npc.position.X, (int)npc.position.Y, (int)npc.width, (int)npc.height );
			Rectangle far = new Rectangle( (int)npc.position.X, (int)npc.position.Y, (int)npc.width, (int)npc.height );

			near.Inflate( 48, 48 );		// 4 blocks
			far.Inflate( 256, 256 );	// 16 blocks

			RandoutLightsOutInPixelUnitRect( near );
			RandoutLightsOutInPixelUnitRect( far );
		}


		public static void RandoutLightsOutInPixelUnitRect( Rectangle pixel_unit_rect ) {
			Vector2 rand_tile_point = RoboUtils.PickRandomTileInPixelUnitRange( pixel_unit_rect );

			int rand_x = (int)rand_tile_point.X;
			int rand_y = (int)rand_tile_point.Y;

			if( Main.tile.GetLength( 0 ) <= rand_x || Main.tile.GetLength( 1 ) <= rand_y ) { return; }

			Tile tile = Main.tile[rand_x, rand_y];
			if( tile != null ) {
				if( tile.type == 4 ) {
//RoboUtils.MsgAll( "torch x:" + rand_x + ", y:" + rand_y + ", style:" + RoboUtils.GetTileStyleTorch( tile ), RoboUtils.MsgType.Debug );
					// Everything except Demon Torches affected
					if( RoboUtils.GetTileStyleTorch( tile ) != 7 ) {
						RoboUtils.TileLightOff( rand_x, rand_y );
					}
				}
			}
		}


		private void UpdateBuffCleanse() {
			Dictionary<int, int> buff_map = RoboUtils.GetBuffs( npc );
			foreach( KeyValuePair<int, int> kv in buff_map ) {
				int buff_type = kv.Key;
				int buff_index = kv.Value;

				// Arrest velocity of lovestruck reaper
				if( RoboUtils.GetBuffs( npc ).ContainsKey( BuffDef.byName["Vanilla:Lovestruck"] ) ) {
					npc.velocity.X = 0;
					npc.velocity.Y = 0;
				} else {
					npc.DelBuff( buff_index );
				}
			}
		}


		private void UpdateStealth() {
			if( PlayerTargets.ContainsKey( this ) ) {
				float distance = 999999;
				FTR_Player ftr_player = PlayerTargets[this];

				Player player = ftr_player.player;
				if( player.active ) {
					distance = Vector2.Distance( player.position, npc.position ) / 16f;
				}

				if( HasStealth ) {
					int threshold = AuraNear;	// (AuraNear + AuraVeryNear) / 2;
					// Too close? End stealth
					if( distance <= threshold ) {
RoboUtils.MsgAll( "Stealth off (far) "+npc.alpha, RoboUtils.MsgType.Debug );
						hasStealth = false;
					}
				} else {
					int threshold = AuraFar;
					// Too far? Reaper not fully enraged? Reaper not in pain? Resume stealth
					if( distance > AuraFar && ftr_player.ReaperRage < 10 && !VelocityClampSkipClock.IsRunning ) {
RoboUtils.MsgAll( "Stealth on (far) "+npc.alpha, RoboUtils.MsgType.Debug );
						hasStealth = true;
					}
				}
			}

			if( HasStealth ) {
				float invis_alpha = 246f;
				if( CurrentAlpha < invis_alpha ) {
					npc.alpha = (int)(CurrentAlpha += 1.25f);
				} else if( CurrentAlpha > invis_alpha ) {
					npc.alpha = (int)(CurrentAlpha -= 1.25f);
				}
			} else {
				if( CurrentAlpha > (float)DefaultAlpha ) {
					npc.alpha = (int)(CurrentAlpha -= 0.75f);
				}
			}
		}


		private void UpdateDeathAura() {
			for( int i = 0; i < Main.npc.Length; i++ ) {
				NPC victim = Main.npc[i];

				if( victim == null || victim.life == 0 || !victim.active ) { continue; }
				//if( victim.townNPC ) { continue; }
				if( victim.whoAmI == npc.whoAmI ) { continue; }
				if( victim.boss ) { continue; }

				ProjectDeathAuraToNPC( victim );
			}
		}

		private bool ProjectDeathAuraToNPC( NPC victim ) {
			float world_distance = Vector2.Distance( victim.position, npc.position );
			float distance = world_distance / 16f;
			int threshold = (AuraNear + AuraVeryNear) / 2; ; // (AuraNear + AuraFar) / 2;
			if( distance > threshold ) { return false; }

			int death_odds = (int)Math.Max( (int)world_distance / threshold, 2 );
			bool is_dead = false;

			if( Main.rand.Next( death_odds ) == 0 ) {
				victim.life = 0;
				is_dead = true;
			}
RoboUtils.MsgAll( "death aura " + victim.whoAmI + ":" + victim.name + ", odds:" + death_odds+", dead:"+is_dead, RoboUtils.MsgType.Debug );

			return is_dead;
		}
	}
}

